/*
 * Glacier - another Gnome email client.
 * Copyright (C) 1999 Lubomir Gelo
 *
 * Glacier uses code from other GPL-ed projects. See THANKS for 
 * credits and copyright info.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "includes.h"

const char RFC822Specials[] = "@.,:;<>[]\\\"()";
#define is_special(x) strchr(RFC822Specials,x)

static Address *address_add_address(const char *phrase, char *comment, size_t *commentlen, size_t commentmax);

static const char *address_parse_comment (const char *s, char *comment, size_t *commentlen, size_t commentmax);
static const char *address_parse_quote (const char *s, char *token, size_t *tokenlen, size_t tokenmax);
static const char *address_parse_mailboxdomain (const char *s, const char *nonspecial, char *mailbox, size_t *mailboxlen, size_t mailboxmax, char *comment, size_t *commentlen, size_t commentmax);
static const char *address_parse_address (const char *s, char *token, size_t *tokenlen, size_t tokenmax, char *comment, size_t *commentlen, size_t commentmax, Address *addr);
static const char *address_parse_route_addr (const char *s, char *comment, size_t *commentlen, size_t commentmax, Address *addr);
static const char *address_parse_addr_spec (const char *s, char *comment, size_t *commentlen, size_t commentmax, Address *addr);

static const char *address_get_next_token (const char *s, char *token, size_t *tokenlen, size_t tokenmax);

/* static void address_list_print_debug(GList *list); */

gchar *address_print_short(GList *list, guint n){

    gchar 	*string = NULL;
    GList	*p;
    Address 	*addr;

    p = g_list_nth(list, n);
    
    if (p){
	addr = (Address *)(p->data);
	if (addr){
	    if (addr->person)
		string = g_strdup(addr->person);
	    else if (addr->mailbox)
		string = g_strdup(addr->mailbox);		
	}
    }
    
    return string;
}


gchar *address_list_print(GList *list){

    gchar 	*string = NULL;
    GList	*p;
    Address 	*addr;
    gint	length = 0, i=0, count = 0;
    gchar	*separator = "\n";
    gint	separator_length;
    
    separator_length = strlen(separator);

    p = g_list_first(list);
    
    while (p){
	addr = (Address *)(p->data);

	if (addr->person)
	    length += strlen(addr->person);
	if (addr->mailbox)
	     length += strlen(addr->mailbox) + 3;
	
	 length += separator_length;
	++count;
	
	p = g_list_next(p);
    }

    string = g_malloc(length+1);
    string[0]='\0';
    
    p = g_list_first(list);
    
    while (p){
	addr = (Address *)(p->data);

	if (addr){
	    gchar *next = address_print(addr);

	    ++i;
	    if (next){
		strcat(string, next);
		g_free(next);
		if (i < count)
		    strcat(string, separator);

	    }
	}
	p = g_list_next(p);
    }

    return string;
}

gchar *address_print(Address *addr){

    if (addr->person){
	
	gint length = strlen(addr->person) + strlen(addr->mailbox) + 4;
	gchar *string = g_malloc(length);

	if (string)
	    snprintf(string, length, "%s <%s>", addr->person, addr->mailbox);

	return string;

    } else {

	return g_strdup(addr->mailbox);
    }
}


Address *address_new(void){

    Address *address;
    
    address = (Address *)g_malloc(sizeof(Address));
    memset(address, 0, sizeof(Address));
    
    return address;
}

void address_list_free(GList *list){

    Address 	*addr;
    GList	*p;

    p = g_list_first(list);
    
    while (p){
	addr = (Address *)(p->data);
	if (addr)
	    address_free(addr);
	
	p = g_list_next(p);
    }

    g_list_free(list);
}

void address_free(Address *address){

    if (!address)
	return;
	
    G_FREE(address->full);
    G_FREE(address->person);
    G_FREE(address->mailbox);
    
    g_free(address);
}

GList *address_parse_addrlist (const char *s){

    const char *begin, *ps;
    char comment[STRING], phrase[STRING];
    size_t phraselen = 0, commentlen = 0;

    Address 	*address;
    GList	*last=NULL;
    
    SKIPWS (s);
    begin = s;

/*    fprintf(stderr,"=>%s\n",s); */

    while (*s) {

	if (*s == ',') {

            if (phraselen) {
		phrase[phraselen] = 0;
		address = address_add_address(phrase, comment, &commentlen, sizeof (comment) - 1);

		if (address){
		    last=g_list_append(last, address);
		    last=g_list_last(last);
		}

            } else if (commentlen && last && !((Address *)(last->data))->person) {

	        comment[commentlen] = 0;
		((Address *)(last->data))->person = g_strdup (comment);
            }

            if (last && !((Address *)(last->data))->full)
		((Address *)(last->data))->full = utils_substrdup (begin, s);

            commentlen = 0;
            phraselen = 0;
            s++;
            begin = s;
            SKIPWS (begin);

	} else if (*s == '(') {

            if (commentlen && commentlen < sizeof (comment) - 1)
		comment[commentlen++] = ' ';

            if ((ps = address_get_next_token (s, comment, &commentlen, sizeof (comment) - 1)) == NULL) {
	        return NULL;
            }

            s = ps;

	} else if (*s == ':') {

            address = address_new();
            phrase[phraselen] = 0;

            address->mailbox = g_strdup (phrase);
            address->group = 1;
            address->full = utils_substrdup (begin, s);

	    if (address){
		last=g_list_append(last, address);
		last=g_list_last(last);
	    }

            phraselen = 0;
            commentlen = 0;
            s++;
            begin = s;
            SKIPWS (begin);

	} else if (*s == ';') {

            if (phraselen) {
		phrase[phraselen] = 0;
		address = address_add_address(phrase, comment, &commentlen, sizeof (comment) - 1);

		if (address){
		    last=g_list_append(last, address);
		    last=g_list_last(last);
		}

            } else if (commentlen && !((Address *)(last->data))->person) {
		comment[commentlen] = 0;
		((Address *)(last->data))->person = g_strdup (comment);
            }

            if (last && !((Address *)(last->data))->full)
		((Address *)(last->data))->full = utils_substrdup (begin, s);

            /* add group terminator */
            address = address_new();

	    if (address){
	        last=g_list_append(last, address);
		last=g_list_last(last);
	    }

            phraselen = 0;
            commentlen = 0;
            s++;
            begin = s;
            SKIPWS (begin);

	} else if (*s == '<') {

            phrase[phraselen] = 0;
            address = address_new();

	    if (address){
	        last=g_list_append(last, address);
		last=g_list_last(last);
	    }

            if (phraselen){

		if (((Address *)(last->data))->person)
	            g_free (((Address *)(last->data))->person);
	
		/* if we get something like "Michael R. Elkins" remove the quotes */
		utils_dequote_comment (phrase);
		((Address *)(last->data))->person = g_strdup (phrase);
            }

            if ((ps = address_parse_route_addr (s + 1, comment, &commentlen, sizeof (comment) - 1, (Address *)(last->data))) == NULL){
		return NULL;
            }


            phraselen = 0;
            commentlen = 0;
            s = ps;

	} else {

            if (phraselen && phraselen < sizeof (phrase) - 1)
		phrase[phraselen++] = ' ';

            if ((ps = address_get_next_token(s, phrase, &phraselen, sizeof (phrase) - 1)) == NULL) {
		return NULL;
            }

            s = ps;
	}

	SKIPWS (s);
    }
  
    if (phraselen) {

	phrase[phraselen] = 0;
	comment[commentlen] = 0;
	address = address_add_address(phrase, comment, &commentlen, sizeof (comment) - 1);

	if (address){
	    last=g_list_append(last, address);
	    last=g_list_last(last);
	}


    } else if (commentlen && last && !((Address *)(last->data))->person) {

	comment[commentlen] = 0;
	((Address *)(last->data))->person = g_strdup (comment);
    }

    if (last)
	((Address *)(last->data))->full = utils_substrdup (begin, s);

/*    address_list_print_debug(last); */
    return g_list_first(last);
}


static const char *address_parse_addr_spec (const char *s, char *comment, size_t *commentlen, size_t commentmax, Address *addr){
    char token[STRING];
    size_t tokenlen = 0;

    s = address_parse_address (s, token, &tokenlen, sizeof (token) - 1, comment, commentlen, commentmax, addr);

    if (s && *s && *s != ',' && *s != ';') {
	return NULL;
    }

    return s;
}

static Address *address_add_address(const char *phrase, char *comment, size_t *commentlen, size_t commentmax){

    Address *address = address_new ();
  
    if (address_parse_addr_spec(phrase, comment, commentlen, commentmax, address) == NULL)
	return NULL;

    return address;
}

static const char *address_parse_comment (const char *s, char *comment, size_t *commentlen, size_t commentmax){

    int level = 1;
  
    while (*s && level){

	if (*s == '(')
    	    level++;
	else if (*s == ')'){

    	    if (--level == 0){
		s++;
		break;
             }
	} else if (*s == '\\') {

            if (!*++s)
		break;
	}

	if (*commentlen < commentmax)
            comment[(*commentlen)++] = *s;

	s++;
    }

    if (level){
	return NULL;
    }

    return s;
}

static const char *address_parse_quote (const char *s, char *token, size_t *tokenlen, size_t tokenmax){

    if (*tokenlen < tokenmax)
	token[(*tokenlen)++] = '"';

    while (*s){

	if (*tokenlen < tokenmax)
    	    token[(*tokenlen)++] = *s;
    
	if (*s == '"')
            return (s + 1);

	if (*s == '\\'){

            if (!*++s)
		break;
	}

	s++;
    }

    return NULL;
}

static const char *address_get_next_token (const char *s, char *token, size_t *tokenlen, size_t tokenmax){

    if (*s == '(')
	return (address_parse_comment (s + 1, token, tokenlen, tokenmax));

    if (*s == '"')
	return (address_parse_quote (s + 1, token, tokenlen, tokenmax));

    if (is_special (*s)) {

	if (*tokenlen < tokenmax)
            token[(*tokenlen)++] = *s;

	return (s + 1);
    }

    while (*s) {
    
	if (ISSPACE (*s) || is_special (*s))
            break;

	if (*tokenlen < tokenmax)
            token[(*tokenlen)++] = *s;
    
	s++;
    }

    return s;
}

static const char *address_parse_mailboxdomain (const char *s, const char *nonspecial, char *mailbox, size_t *mailboxlen, size_t mailboxmax, char *comment, size_t *commentlen, size_t commentmax){

    const char *ps;

    while (*s){

	SKIPWS (s);
	if (strchr (nonspecial, *s) == NULL && is_special (*s))
            return s;

	if (*s == '(') {

            if (*commentlen && *commentlen < commentmax)
		comment[(*commentlen)++] = ' ';

            ps = address_get_next_token (s, comment, commentlen, commentmax);

	} else
            ps = address_get_next_token (s, mailbox, mailboxlen, mailboxmax);

	if (!ps)
            return NULL;
    
	s = ps;
    }

    return s;
}

static const char *address_parse_address (const char *s, char *token, size_t *tokenlen, size_t tokenmax, char *comment, size_t *commentlen, size_t commentmax, Address *addr){

    s = address_parse_mailboxdomain (s, ".\"(\\", token, tokenlen, tokenmax, comment, commentlen, commentmax);

    if (!s)
	return NULL;

    if (*s == '@') {

	if (*tokenlen < tokenmax)
            token[(*tokenlen)++] = '@';

	s = address_parse_mailboxdomain (s + 1, ".([]\\", token, tokenlen, tokenmax, comment, commentlen, commentmax);

	if (!s)
            return NULL;
    }

    token[*tokenlen] = 0;
    addr->mailbox = g_strdup (token);

    if (*commentlen && !addr->person) {
	comment[*commentlen] = 0;
	addr->person = g_strdup (comment);
    }

    return s;
}

static const char *address_parse_route_addr (const char *s, char *comment, size_t *commentlen, size_t commentmax, Address *addr){

    char token[STRING];
    size_t tokenlen = 0;

    SKIPWS (s);

    /* find the end of the route */
    if (*s == '@'){

	while (s && *s == '@'){
    
    	    if (tokenlen < sizeof (token) - 1)
		token[tokenlen++] = '@';

            s = address_parse_mailboxdomain (s + 1, ".\\[](", token, &tokenlen, sizeof (token) - 1, comment, commentlen, commentmax);
	}

	if (!s || *s != ':') {
            return NULL; /* invalid route */
	}

	if (tokenlen < sizeof (token) - 1)
            token[tokenlen++] = ':';
	    s++;
    }

    if ((s = address_parse_address (s, token, &tokenlen, sizeof (token) - 1, comment, commentlen, commentmax, addr)) == NULL)
	return NULL;

    if (*s != '>' || !addr->mailbox){
	return NULL;
    }

    s++;
    return s;
}

Address *address_copy(Address *addr){

    Address *p = address_new();

    p->full = g_strdup (addr->full);
    p->person = g_strdup (addr->person);
    p->mailbox = g_strdup (addr->mailbox);
    p->group = addr->group;

    return p;
}

GList *address_list_copy(GList *list){

    GList 	*p = g_list_first(list);
    GList	*new_list = NULL;
    Address	*addr;
    
    while (p){
	addr = address_copy((Address *)(p->data));
	if (addr)
	    new_list = g_list_append(new_list, addr);
	
	p = g_list_next(p);
    }

    return new_list;
}


/*
static void address_list_print_debug(GList *list){
    
    GList 	*p;
    Address 	*addr;

    if (!list){
	fprintf(stderr, "ADDRESS DEBUG: list is NULL!\n");
    }
    
    p = g_list_first(list);

    while (p){

	addr = (Address *)(p->data);

	fprintf(stderr,"->");

	if (addr->full)
	    fprintf(stderr, "F: %s ", addr->full);

	if (addr->person)
	    fprintf(stderr, "P: %s ", addr->person);

	if (addr->mailbox)
	    fprintf(stderr, "M: %s ", addr->mailbox);

	fprintf(stderr,"\n");

	p = g_list_next(p);
    }

}

*/





