/*
 * Glacier - another Gnome email client.
 * Copyright (C) 1999 Lubomir Gelo
 *
 * Glacier uses code from other GPL-ed projects. See THANKS for 
 * credits and copyright info.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "includes.h"

static int mbox_load_messages(Mailbox *mailbox);
static int mbox_lock_mailbox (Mailbox *mailbox, int exclusive, int retries);
static void mbox_unlock_mailbox (Mailbox *mailbox);
static time_t mbox_is_from (const char *s, char *path, size_t pathlen);

int mbox_open_mailbox (Mailbox *mailbox, gboolean read_messages){

    int ret;

    if ((mailbox->fp = fopen (mailbox->path, "r")) == NULL){
	return (-1);
    }

    utils_block_signals();

    if (mbox_lock_mailbox(mailbox, 0, 1) == -1){
        utils_unblock_signals();
        return (-1);
    }

    if (read_messages)
        ret = mbox_load_messages(mailbox);
    else
	ret = 0;	

    mbox_unlock_mailbox(mailbox);
    utils_unblock_signals();

    return (ret);
}


int mbox_close_mailbox (Mailbox *mailbox) {

    gint 	i;
    Message 	*message;

    utils_unlock_file(mailbox->path, fileno(mailbox->fp));
    utils_unblock_signals();

    if (mailbox->fp)
	fclose(mailbox->fp);

    mailbox->fp = NULL;	

    for(i = 0; i < mailbox->total_messages; i++) {
	message = (Message *)(g_list_last(mailbox->messages)->data);
        mailbox->messages = g_list_remove(mailbox->messages, message);
	message_free(message);
    }

    mailbox->total_messages = 0;
    mailbox->new_messages = 0;
	
    return 0;
}


static void mbox_unlock_mailbox (Mailbox *mailbox){

    if (mailbox->locked){
	fflush(mailbox->fp);

	utils_unlock_file(mailbox->path, fileno (mailbox->fp));
        mailbox->locked = 0;
    }
}


static int mbox_lock_mailbox(Mailbox *mailbox, int exclusive, int retries){

    int ret;

    if ((ret = utils_lock_file(mailbox->path, fileno (mailbox->fp), exclusive, 1, retries)) == 0)
	mailbox->locked = 1;

    return (ret);
}

static int mbox_load_messages(Mailbox *mailbox){

    struct stat sb;
    char buf[HUGE_STRING];
    char return_path[STRING];
    Message *msg;
    time_t t, tz;
    int count = 0, lines = 0, new_msgs = 0;
    long loc;
    int i = 0;

#ifdef NFS_ATTRIBUTE_HACK
    struct utimbuf newtime;
#endif

    /* Save information about the folder at the time we opened it. */
    if (stat (mailbox->path, &sb) == -1){
	return (-1);
    }

    mailbox->size = sb.st_size;
    mailbox->mtime = sb.st_mtime;

#ifdef NFS_ATTRIBUTE_HACK
    if (sb.st_mtime > sb.st_atime){
	newtime.modtime = sb.st_mtime;
	newtime.actime = time (NULL);
	utime (mailbox->path, &newtime);
    }
#endif

    tz = utils_local_tz ();

    if (!mailbox->readonly)
	mailbox->readonly = (gboolean)access(mailbox->path, W_OK) ? 1 : 0;

    loc = ftell (mailbox->fp);

    while (fgets (buf, sizeof (buf), mailbox->fp) != NULL){

	if ((t = mbox_is_from (buf, return_path, sizeof (return_path)))) {

	    if (count > 0){
	
		Message *last = (Message *)((g_list_last(mailbox->messages))->data);

		if (MSG_CONTENT(last)->length < 0){
		    
		    MSG_CONTENT(last)->length = loc - MSG_CONTENT(last)->offset - 1;

		    if (MSG_CONTENT(last)->length < 0)
			MSG_CONTENT(last)->length = 0;

		}
	
		if (!last->header->lines)
	            last->header->lines = lines ? lines - 1 : 0;
	    }


	    msg = message_new(mailbox);
	    message_read_header(mailbox->fp, msg);
     
            msg->header->received = t + tz;
	    if (msg->header->date_sent == 0)
		msg->header->date_sent = msg->header->received;

            msg->header->index = count;

            if (!msg->envelope->return_path && return_path[0]){
		 msg->envelope->return_path = address_parse_addrlist(return_path);
	    }

            if (!msg->envelope->from){
		msg->envelope->from = address_list_copy(msg->envelope->return_path);
	    }

	    mailbox->messages = g_list_append(mailbox->messages, msg);


            /* if we know how long this message is, either just skip over the body,
             * or if we don't know how many lines there are, count them now (this will
             * save time by not having to search for the next message marker).
             */
	     
            if (MSG_CONTENT(msg)->length > 0){

	        long tmploc;

	        loc = ftell (mailbox->fp);
	        tmploc = loc + MSG_CONTENT(msg)->length + 1;

	        if (tmploc < mailbox->size){
	            /*
	             * check to see if the content-length looks valid.  we expect to
	             * to see a valid message separator at this point in the stream
	             */
	            fseek (mailbox->fp, tmploc, 0);

	            if (fgets (buf, sizeof (buf), mailbox->fp) == NULL || strncmp ("From ", buf, 5) != 0) {
		        fprintf (stderr, "mbox_load_messages: bad content-length cl=%ld\n", MSG_CONTENT(msg)->length);
		        fprintf (stderr, "\tLINE: %s", buf);
		        fseek (mailbox->fp, loc, 0); 		/* nope, return the previous position */
			MSG_CONTENT(msg)->length = -1;
	            }

		} else if (tmploc != mailbox->size) {

	             /* content-length would put us past the end of the file, so it must be wrong */
	             MSG_CONTENT(msg)->length = -1;
	        }

	        if (MSG_CONTENT(msg)->length != -1){
		    
	            /* good content-length.  check to see if we know how many lines* are in this message */
	            if (msg->header->lines == 0) {

		        int cl = MSG_CONTENT(msg)->length;

		        /* count the number of lines in this message */
		        fseek (mailbox->fp, loc, 0);
	    
		        while (cl-- > 0) {
	                    if (fgetc (mailbox->fp) == '\n')
			        msg->header->lines++;
	                }
	            }

	            /* return to the offset of the next message separator */
		    fseek (mailbox->fp, tmploc, 0);
		}
	    }

	    lines = 0;

	    count++;
	    
	    if (! msg->header->read)
		++ new_msgs;

            if(count > i){
    		utils_gui_update_hook(count, ftell (mailbox->fp) / (mailbox->size / 100 + 1));
    		i+=10;
            }

	} else
            lines++;
    
	loc = ftell (mailbox->fp);
    }
  
    if (count > 0){

	Message *last = (Message *)((g_list_last(mailbox->messages))->data);

	if (MSG_CONTENT(last)->length < 0){
		    
	    MSG_CONTENT(last)->length = loc - MSG_CONTENT(last)->offset - 1;

	    if (MSG_CONTENT(last)->length < 0)
		MSG_CONTENT(last)->length = 0;

	}
	
	if (!last->header->lines)
            last->header->lines = lines ? lines - 1 : 0;
    }

    mailbox->messages = g_list_first(mailbox->messages);
    mailbox->total_messages = count;
    mailbox->new_messages = new_msgs;
    return count;
}

/* A valid message separator looks like: From [ <return-path> ] <weekday> <month> <day> <time> [ <timezone> ] <year> */
static time_t mbox_is_from (const char *s, char *path, size_t pathlen){

    struct tm tm;
    int yr;

    if (path)
	*path = 0;

    if (strncmp ("From ", s, 5) != 0)
	return 0;

    s = utils_get_next_word (s); /* skip over the From part. */

    if (!*s)
	return 0;

//    fprintf(stderr, "is_from(): parsing: %s", s);

    if (!utils_is_day_name (s)){

	const char *p;
	size_t len;

	/* looks like we got the return-path, so extract it  */

	if (*s == '"') {
    	    /* sometimes we see bogus addresses like
             *	From "/foo/bar baz/"@dumbdar.com Sat Nov 22 15:29:32 PST 1997
             */
            p = s;
            p++; /* skip over the quote */

            do {

		if (!(p = strpbrk (p, "\\\"")))
	    	    return 0;

		if (*p == '\\')
	            p += 2;
            } while (*p != '"');

            while (*p && !ISSPACE (*p))
		p++;

	} else {

            if ((p = strchr (s, ' ')) == NULL)
		return 0;
	}

	if (path) {

    	    len = (size_t) (p - s);

            if (len + 1 > pathlen)
		len = pathlen - 1;

            memcpy (path, s, len);
            path[len] = 0;
	}

	s = p + 1;
	SKIPWS (s);

	if (!*s)
            return 0;

	if (!utils_is_day_name (s)) {
//            fprintf(stderr, "is_from():  expected weekday, got: %s\n", s);
            return 0;
	}
    }

    s = utils_get_next_word (s);

    if (!*s) 
	return 0;

    /* do a quick check to make sure that this isn't really the day of the week.
     * this could happen when receiving mail from a local user whose login name
     * is the same as a three-letter abbreviation of the day of the week.
     */

    if (utils_is_day_name (s)) {
	s = utils_get_next_word (s);

	if (!*s) 
	    return 0;
    }

    /* now we should be on the month. */
    if ((tm.tm_mon = utils_check_month (s)) < 0) 
	return 0;

    /* day */
    s = utils_get_next_word (s);
    if (!*s) 
	return 0;

    if (sscanf (s, "%d", &tm.tm_mday) != 1) 
	return 0;

    /* time */
    s = utils_get_next_word (s);
    if (!*s) 
	return 0;

    /* Accept either HH:MM or HH:MM:SS */
    if (sscanf (s, "%d:%d:%d", &tm.tm_hour, &tm.tm_min, &tm.tm_sec) == 3)
	;
    else if (sscanf (s, "%d:%d", &tm.tm_hour, &tm.tm_min) == 2)
	tm.tm_sec = 0;
    else
	return 0;

    s = utils_get_next_word (s);
    if (!*s) 
	return 0;

    /* timezone? */
    if (isalpha ((unsigned char) *s) || *s == '+' || *s == '-'){
	s = utils_get_next_word (s);

	if (!*s) 
	    return 0;

	/*
         * some places have two timezone fields after the time, e.g.
         *      From xxxx@yyyyyyy.fr Wed Aug  2 00:39:12 MET DST 1995
         */
    
	if (isalpha ((unsigned char) *s)) {
            s = utils_get_next_word (s);
            if (!*s) 
		return 0;
	}
    }

    /* year */
    if (sscanf (s, "%d", &yr) != 1) 
	return 0;

    tm.tm_year = yr > 1900 ? yr - 1900 : yr;

//    fprintf(stderr, "is_from(): month=%d, day=%d, hr=%d, min=%d, sec=%d, yr=%d.\n", tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, tm.tm_year);

    tm.tm_isdst = 0;
    tm.tm_yday = 0;
    tm.tm_wday = 0;

    return (utils_mktime (&tm, 0));
}

gchar *mbox_get_raw_message(Message *msg){

    Mailbox 	*mbox;
    gchar	*buffer = NULL;
    gchar	line[HUGE_STRING];
    GList	*list;
    long	size, next_offset, loaded;

    if (!msg || msg->mailbox == NULL)
	return NULL;

    mbox = msg->mailbox;
    list = g_list_find(mbox->messages, (gpointer)msg);

    if (!list)
	return NULL;
	
    list = g_list_next(list);
    
    if (!list)
	next_offset = msg->mailbox->size;
    else {
	next_offset = ((Message *)(list->data))->header->offset;
    }	

    size = next_offset - msg->header->offset + 1;
    buffer = g_malloc(size); 
    buffer[0] = '\0';

    fseek(mbox->fp, msg->header->offset, SEEK_SET);

    loaded = 0L;
    while (loaded < size){

	if (fgets (line, sizeof (line), mbox->fp) != NULL){
	    
	    loaded += strlen(line);

	    if (loaded < size){
    	        if (strncmp("From ", line, 5)){
		    strcat(buffer, line);
		} else {
		    break;
		}    
	    }
	} else
	    break;
    }

    return buffer;
}



int mbox_quick_scan(Mailbox *mailbox){

    struct 	stat sb;
    char 	buf[HUGE_STRING];
    char 	return_path[STRING];
    int 	count = 0, new_msgs = 0;

#ifdef NFS_ATTRIBUTE_HACK
    struct utimbuf newtime;
#endif

    /* Save information about the folder at the time we opened it. */
    if (stat (mailbox->path, &sb) == -1){
	return (-1);
    }

    mailbox->size = sb.st_size;
    mailbox->mtime = sb.st_mtime;

#ifdef NFS_ATTRIBUTE_HACK
    if (sb.st_mtime > sb.st_atime){
	newtime.modtime = sb.st_mtime;
	newtime.actime = time (NULL);
	utime (mailbox->path, &newtime);
    }
#endif

    if ((mailbox->fp = fopen (mailbox->path, "r")) == NULL){
	return (-1);
    }

    utils_block_signals();

    if (mbox_lock_mailbox(mailbox, 0, 1) == -1){
        utils_unblock_signals();
        return (-1);
    }

    while (fgets (buf, sizeof (buf), mailbox->fp) != NULL){

	if (mbox_is_from (buf, return_path, sizeof (return_path))) {

	    ++count;
	    ++new_msgs;

   	if (message_check_read_status(mailbox->fp) == 1)		/* is new ? */
		--new_msgs;

	}
    }

    mailbox->total_messages = count;
    mailbox->new_messages = new_msgs;

    mbox_unlock_mailbox(mailbox);
    utils_unblock_signals();

    if (mailbox->fp)
	fclose(mailbox->fp);

    mailbox->fp = NULL;	

    return count;
}
