/*
 * Glacier - another Gnome email client.
 * Copyright (C) 1999 Lubomir Gelo
 *
 * Glacier uses code from other GPL-ed projects. See THANKS for 
 * credits and copyright info.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "includes.h"

static struct tz_t {
    char *tzname;
    unsigned char zhours;
    unsigned char zminutes;
    unsigned char zoccident; /* west of UTC? */
    unsigned char xxx;       /* unused */
}

TimeZones[] = {
    { "sst",  11,  0, 1, 0 }, /* Samoa */
    { "pst",   8,  0, 1, 0 },
    { "mst",   7,  0, 1, 0 },
    { "pdt",   7,  0, 1, 0 },
    { "cst",   6,  0, 1, 0 },
    { "mdt",   6,  0, 1, 0 },
    { "cdt",   5,  0, 1, 0 },
    { "est",   5,  0, 1, 0 },
    { "ast",   4,  0, 1, 0 }, /* Atlantic */
    { "edt",   4,  0, 1, 0 },
    { "wgt",   3,  0, 1, 0 }, /* Western Greenland */
    { "wgst",  2,  0, 1, 0 }, /* Western Greenland DST */
    { "aat",   1,  0, 1, 0 }, /* Atlantic Africa Time */
    { "egt",   1,  0, 1, 0 }, /* Eastern Greenland */
    { "egst",  0,  0, 0, 0 }, /* Eastern Greenland DST */
    { "gmt",   0,  0, 0, 0 },
    { "utc",   0,  0, 0, 0 },
    { "wat",   0,  0, 0, 0 }, /* West Africa */
    { "wet",   0,  0, 0, 0 }, /* Western Europe */
    { "bst",   1,  0, 0, 0 }, /* British DST */
    { "cat",   1,  0, 0, 0 }, /* Central Africa */
    { "cet",   1,  0, 0, 0 }, /* Central Europe */
    { "met",   1,  0, 0, 0 }, /* this is now officially CET */
    { "west",  1,  0, 0, 0 }, /* Western Europe DST */
    { "cest",  2,  0, 0, 0 }, /* Central Europe DST */
    { "eet",   2,  0, 0, 0 }, /* Eastern Europe */
    { "ist",   2,  0, 0, 0 }, /* Israel */
    { "sat",   2,  0, 0, 0 }, /* South Africa */
    { "ast",   3,  0, 0, 0 }, /* Arabia */
    { "eat",   3,  0, 0, 0 }, /* East Africa */
    { "eest",  3,  0, 0, 0 }, /* Eastern Europe DST */
    { "idt",   3,  0, 0, 0 }, /* Israel DST */
    { "msk",   3,  0, 0, 0 }, /* Moscow */
    { "adt",   4,  0, 0, 0 }, /* Arabia DST */
    { "msd",   4,  0, 0, 0 }, /* Moscow DST */
    { "gst",   4,  0, 0, 0 }, /* Presian Gulf */
    { "smt",   4,  0, 0, 0 }, /* Seychelles */
    { "ist",   5, 30, 0, 0 }, /* India */
    { "ict",   7,  0, 0, 0 }, /* Indochina */
/*  { "cst",   8,  0, 0, 0 },*/ /* China */
    { "hkt",   8,  0, 0, 0 }, /* Hong Kong */
/*  { "sst",   8,  0, 0, 0 },*/ /* Singapore */
    { "wst",   8,  0, 0, 0 }, /* Western Australia */
    { "jst",   9,  0, 0, 0 }, /* Japan */
/*  { "cst",   9, 30, 0, 0 },*/ /* Australian Central Standard Time */
    { "kst",  10,  0, 0, 0 }, /* Korea */
    { "nzst", 12,  0, 0, 0 }, /* New Zealand */
    { "nzdt", 13,  0, 0, 0 }, /* New Zealand DST */
    { NULL,    0,  0, 0, 0 }
};


static char *message_read_line (FILE *f, char *line, size_t *linelen);
static time_t message_parse_date (char *s, Header *h);
static char *message_extract_message_id (const char *s);
static GList *message_parse_references (char *s);
static void message_parse_content_disposition (char *s, Body *content);
int message_check_mime_type (const char *s);
static void message_parse_content_type (char *s, Body *content); 
int message_check_encoding (const char *c);
GList *message_parse_message(FILE *fp, Body *parent);
GList *message_parse_multipart(FILE *fp, const char *boundary, long end_off, int digest);
Body *message_read_mime_header (FILE *fp, int digest);

void message_parse_mime_message (Message *msg);


gchar *parameter_get (const char *s, GList *list);

static GList *parameters_parse(const char *s);
static int message_is_text_type (int t, char *s);

Message *message_new(Mailbox *mailbox) {

    Message *message;

    message = (Message *) g_malloc (sizeof (Message));
    memset(message, 0, sizeof(Message));

    message->mailbox = mailbox;

    return message;
}

void message_free(Message *message){

    if (!message)
	return;

    if (message->envelope)
	envelope_free(message->envelope);

    if (message->header)
	header_free(message->header);

    if (message->content)
	body_free(message->content);

    g_free(message);
}

Body *body_new(void){
    
    Body *body;
    
    body = (Body *)g_malloc(sizeof(Body));
    memset(body, 0, sizeof(Body));
    
    return body;
}

void body_free(Body *body){

    g_free(body);
}


Parameter *parameter_new(void){
    
    Parameter *param;
    
    param = (Parameter *)g_malloc(sizeof(Parameter));
    memset(param, 0, sizeof(Parameter));
    
    return param;
}

void parameter_free(Parameter *param){

    G_FREE(param->attribute);
    G_FREE(param->value);
    g_free(param);
}

void parameters_free(GList *list){

    GList *l = g_list_first(list);

    while (l){
        parameter_free((Parameter *)(l->data));
	l = g_list_next(l);
    }

    g_list_free(g_list_first(list));
}

static GList *parameters_parse(const char *s){

    const gchar 	*p;
    GList		*list = NULL;
    char 		buffer[LONG_STRING];
    Parameter		*param;
    size_t 		i;

    while (*s){
	if ((p = strpbrk (s, "=;")) == NULL) {
            dprint(1, (stderr, "parameters_parse(): malformed parameter: %s\n", s));
	    return(g_list_first(list));
	}

	/* if we hit a ; now the parameter has no value, just skip it */
	if (*p != ';'){
    	    i = p - s;

            param = parameter_new();
	    param->attribute = g_malloc (i + 1);
            memcpy(param->attribute, s, i);
            param->attribute[i] = 0;

	    /* remove whitespace from the end of the attribute name */
            while (ISSPACE (param->attribute[--i]))
    		param->attribute[i] = 0;

            s = p + 1; /* skip over the = */
	    SKIPWS (s);

            if (*s == '"'){
	
	        s++;
		for (i=0; *s && *s != '"' && i < sizeof (buffer) - 1; i++, s++){

		    if (*s == '\\'){
			/* Quote the next character */
    			buffer[i] = s[1];
			if (!*++s)
	            	    break;
	    	    } else
		       buffer[i] = *s;
		}

		buffer[i] = 0;

		if (*s)
	    	    s++; /* skip over the " */
	    } else {

		for (i=0; *s && *s != ' ' && *s != ';' && i < sizeof (buffer) - 1; i++, s++)
	    	    buffer[i] = *s;

		buffer[i] = 0;
    	    }

    	    param->value = g_strdup(buffer);
	    list = g_list_append(list, param);

	} else {
    	    s = p;
            dprint (1, (stderr, "parameters_parse(): parameter with no value: %s\n", s));
	}

	/* Find the next parameter */
	if (*s != ';' && (s = strchr (s, ';')) == NULL)
	    break; /* no more parameters */

	do {
            s++;

            /* Move past any leading whitespace */
            SKIPWS (s);
	} while (*s == ';'); /* skip empty parameters */
    }    

    return(g_list_first(list));
}

gchar *parameter_get (const char *s, GList *list){

    Parameter *p;
    GList *l = g_list_first(list);
    
    while (l){
	
	p = (Parameter *)(l->data);
	if (strcasecmp(s, p->attribute) == 0)
	    return p->value;

	l = g_list_next(l);
    }

    return NULL;
}



Envelope *envelope_new(void){

    Envelope *envelope;
    
    envelope = (Envelope *)g_malloc (sizeof(Envelope));
    memset(envelope, 0, sizeof(Envelope));
    
    return envelope;
}

void envelope_free(Envelope *envelope){

    GList *list;

    if (!envelope)
	return;
	
    if (envelope->return_path)
	address_list_free(envelope->return_path);

    if (envelope->from)
	address_list_free(envelope->from);

    if (envelope->to)
	address_list_free(envelope->to);

    if (envelope->cc)
	address_list_free(envelope->cc);

    if (envelope->bcc)
	address_list_free(envelope->bcc);

    if (envelope->sender)
	address_list_free(envelope->sender);

    if (envelope->reply_to)
	address_list_free(envelope->reply_to);

    if (envelope->mail_followup_to)
	address_list_free(envelope->mail_followup_to);

    if (envelope->references){
	list = g_list_first(envelope->references);

	while (list){
	    G_FREE(list->data);
	    list = g_list_next(list);
	}

	g_list_free(envelope->references);
    }

    G_FREE(envelope->subject);
    G_FREE(envelope->message_id);
    G_FREE(envelope->supersedes);
    
    g_free(envelope);
}

Header *header_new(void){

    Header *header;
    
    header = (Header *)g_malloc (sizeof(Header));
    memset(header, 0, sizeof(Header));
    
    return header;
}

void header_free(Header *header){

    if (header)
	g_free(header);

}


void message_read_header (FILE *file, Message *msg){

    long 	current_loc;
    char 	*line = g_malloc (LONG_STRING);
    char 	*p;
    int 	matched;
    size_t 	linelen = LONG_STRING;
    char 	in_reply_to[STRING];

    in_reply_to[0] = 0;

    current_loc = ftell(file);

    if (msg->content == NULL){
	msg->content = body_new();
    }

    if (msg->envelope == NULL)
	msg->envelope = envelope_new();

    if (msg->header == NULL)
	msg->header = header_new();

    msg->header->offset = current_loc;		/* where this message start */

    MSG_CONTENT(msg)->type = TYPETEXT;
    MSG_CONTENT(msg)->subtype = g_strdup ("plain");
    MSG_CONTENT(msg)->encoding = ENC7BIT; 
    MSG_CONTENT(msg)->length = -1;

    while (*(line = message_read_line (file, line, &linelen)) != 0) {

	matched = 0;

	if ((p = strpbrk (line, ": \t")) == NULL || *p != ':') {

            // some bogus MTAs will quote the original "From " line
            if (strncmp (">From ", line, 6) == 0) {
		current_loc = ftell (file);
	        continue; // just ignore 
            }

            fseek (file, current_loc, 0);
            break; // end of header
	}

	*p = 0;
	p++;
	SKIPWS (p);

	if (!*p){
	    current_loc = ftell(file);
            continue; // skip empty header fields
	}

	switch (tolower (line[0])) {
      
            case 'a':
		if (strcasecmp (line+1, "pparently-to") == 0){
	            msg->envelope->to = address_parse_addrlist(p);
	            matched = 1;
		} else if (strcasecmp (line+1, "pparently-from") == 0){
	            msg->envelope->from = address_parse_addrlist(p);
	            matched = 1;
	        }
		break;

	    case 'b':
		if (strcasecmp (line+1, "cc") == 0){
	            msg->envelope->bcc = address_parse_addrlist(p);
	            matched = 1;
		}
		break;

	    case 'c':
		if (strcasecmp (line+1, "c") == 0){
	            msg->envelope->cc = address_parse_addrlist(p);
	            matched = 1;

		}  else if (strncasecmp (line + 1, "ontent-", 7) == 0){

        	    if (strcasecmp (line+8, "type") == 0) {

		        message_parse_content_type (p, MSG_CONTENT(msg));
			matched = 1;					

		    } else if (strcasecmp (line+8, "transfer-encoding") == 0) {
	    
	                msg->content->encoding = message_check_encoding (p);
    			matched = 1;					

	            } else if (strcasecmp (line+8, "length") == 0) {

	                msg->content->length = atoi (p);
			matched = 1;

	            } else if (strcasecmp (line+8, "description") == 0) {

	                g_free (msg->content->description);
	                msg->content->description = g_strdup (p);
	                rfc2047_decode (msg->content->description, msg->content->description, strlen (msg->content->description) + 1); 
			matched = 1;

	            } else if (strcasecmp (line+8, "disposition") == 0) {
	    
	                message_parse_content_disposition (p, msg->content);
			matched = 1; 
	            }
		}
		break;

	    case 'd':
		if (!strcasecmp ("ate", line + 1)){

		    msg->header->date_sent = message_parse_date (p, msg->header);
	            matched = 1;
		}
		break;

            case 'e':
		if (!strcasecmp ("xpires", line + 1) &&  message_parse_date (p, msg->header) < time (NULL)){
	            msg->header->expired = 1;
		    matched = 1;
		}
	        break;

            case 'f':
		if (!strcasecmp ("rom", line + 1)){
	            msg->envelope->from = address_parse_addrlist(p);
	            matched = 1;
		}
		break;

            case 'i':
		if (!strcasecmp (line+1, "n-reply-to")){

		    strfcpy (in_reply_to, p, sizeof (in_reply_to));
		    rfc2047_decode (in_reply_to, in_reply_to, sizeof (in_reply_to));
		}
		break;

            case 'l':
		if (!strcasecmp (line + 1, "ines")){
	
		    msg->header->lines = atoi(p);
                    matched = 1;
		}
		break;

            case 'm':
		if (!strcasecmp (line + 1, "ime-version")){

		    msg->header->mime = 1;
	            matched = 1;

		} else if (!strcasecmp (line + 1, "essage-id"))	{
	  
	            // We add a new "Message-Id:" when building a message
	            G_FREE(msg->envelope->message_id);
	            msg->envelope->message_id = message_extract_message_id (p);
	            matched = 1;

		} else if (!strncasecmp (line + 1, "ail-", 4)){

	            if (!strcasecmp (line + 5, "reply-to")) {

			// override the Reply-To: field 
			if (msg->envelope->reply_to)
			    address_list_free(msg->envelope->reply_to);
			
			msg->envelope->reply_to = address_parse_addrlist(p);
			matched = 1;

	    	    } else if (!strcasecmp (line + 5, "followup-to")) {

	    		msg->envelope->mail_followup_to = address_parse_addrlist (p);
			matched = 1;
	            }
		}
		break;


	    case 'r':
		if (!strcasecmp (line + 1, "eferences")){

	            msg->envelope->references = message_parse_references (p);
	            matched = 1;

		} else if (!strcasecmp (line + 1, "eply-to")) {

	            msg->envelope->reply_to = address_parse_addrlist(p);
	            matched = 1;

		} else if (!strcasecmp (line + 1, "eturn-path")){

	            msg->envelope->return_path = address_parse_addrlist(p);
	            matched = 1;

		}  else if (!strcasecmp (line + 1, "eceived")) {
	  
	            if (!msg->header->received) {

			char *d = strchr (p, ';');

			if (d)
	                    msg->header->received = message_parse_date (d + 1, NULL);
	            }
	            matched = 1;
		}
		break;

	    case 's':
		if (!strcasecmp (line + 1, "ubject")){

	            if (!msg->envelope->subject)
			msg->envelope->subject = g_strdup (p);

	            matched = 1;

		} else if (!strcasecmp (line + 1, "ender")) {

	            msg->envelope->sender = address_parse_addrlist(p);
	            matched = 1;

		}  else if (!strcasecmp (line + 1, "tatus")) {

		    while (*p){
	      
	                switch(*p){
			    case 'r':
		                msg->header->replied = 1;
		                break;
		
			    case 'O':
				msg->header->old = 1;
		                break;

			    case 'R':
		                msg->header->read = 1;
		                break;
	                }

	                p++;
		    }
	            matched = 1;

		} else if ((!strcasecmp ("upersedes", line + 1) || !strcasecmp ("upercedes", line + 1))){
	              msg->envelope->supersedes = g_strdup (p);
		      matched = 1;
		}

		break;

	    case 't':
		if (strcasecmp (line+1, "o") == 0){
	            msg->envelope->to = address_parse_addrlist (p);
	            matched = 1;
		}
		break;
		
	    case 'x':
		if (strcasecmp (line+1, "-status") == 0){

		    while (*p){
	            	switch (*p) {
			    case 'A':
		        	msg->header->replied = 1;
		                break;

			    case 'D':
		        	msg->header->deleted = 1;
		                break;

			    case 'F':
		                msg->header->flagged = 1;
		                break;
		
			    default:
		                break;
	                }
	                p++;
		    }

	            matched = 1;
		}
		break;
		

	    default:
		break;
	}

	if (!matched){

            line[strlen (line)] = ':';
	    g_list_append(msg->envelope->userhdrs, g_strdup(line));

	}

	current_loc = ftell(file);
    }

    g_free(line);

    msg->content->offset = current_loc;


    /* do RFC2047 decoding */
    rfc2047_decode_addrlist (msg->envelope->from);
    rfc2047_decode_addrlist (msg->envelope->to);
    rfc2047_decode_addrlist (msg->envelope->cc);
    rfc2047_decode_addrlist (msg->envelope->reply_to);
    rfc2047_decode_addrlist (msg->envelope->mail_followup_to);
    rfc2047_decode_addrlist (msg->envelope->return_path);
    rfc2047_decode_addrlist (msg->envelope->sender);



    if (msg->envelope->subject) {
    	rfc2047_decode (msg->envelope->subject, msg->envelope->subject, strlen (msg->envelope->subject) + 1);
        msg->envelope->real_subj = msg->envelope->subject;
    }

    /* check for missing or invalid date */
    if (msg->header->date_sent <= 0)  {
	    msg->header->date_sent = msg->header->received;
    }


}



static char *message_read_line (FILE *f, char *line, size_t *linelen){

    char *buf = line;
    char ch;
    size_t offset = 0;

    while (1) {
    
	if (fgets (buf, *linelen - offset, f) == NULL || (ISSPACE (*line) && !offset))	{
    	    *line = 0;
            return (line);
	}

	buf += strlen (buf) - 1;

	if (*buf == '\n') {

    	    /* we did get a full line. remove trailing space */
            while (ISSPACE (*buf))
		*buf-- = 0;	/* we cannot come beyond line's beginning because
			         * it begins with a non-space */

            /* check to see if the next line is a continuation line */
            if ((ch = fgetc (f)) != ' ' && ch != '\t') {
		ungetc (ch, f);
		return (line); /* next line is a separate header field or EOH */
            }

            /* eat tabs and spaces from the beginning of the continuation line */
            while ((ch = fgetc (f)) == ' ' || ch == '\t')
		;

            ungetc (ch, f);
            *++buf = ' '; /* string is still terminated because we removed
		             at least one whitespace char above */
	}

	buf++;
	offset = buf - line;

	if (*linelen < offset + STRING) {

    	    /* grow the buffer */
    	    *linelen += STRING;
            g_realloc ((void **) &line, *linelen);
            buf = line + offset;
	}
    }

    return line;
}


static time_t message_parse_date (char *s, Header *h){

    int count = 0;
    char *t;
    int hour, min, sec;
    struct tm tm;
    int i;
    int tz_offset = 0;
    int zhours = 0;
    int zminutes = 0;
    int zoccident = 0;
    const char *ptz;
    char tzstr[SHORT_STRING];

    /* kill the day of the week, if it exists. */
    if ((t = strchr (s, ',')))
	t++;
    else
	t = s;

    SKIPWS (t);

    memset (&tm, 0, sizeof (tm));

    while ((t = strtok (t, " \t")) != NULL){

	switch (count){

    	    case 0: /* day of the month */
		if (!isdigit ((unsigned char) *t))
	            return (-1);
		tm.tm_mday = atoi (t);

		if (tm.tm_mday > 31)
	            return (-1);

		break;

            case 1: /* month of the year */
		
		if ((i = utils_check_month (t)) < 0)
	    	    return (-1);
	
		tm.tm_mon = i;
		break;

            case 2: /* year */

		tm.tm_year = atoi (t);

		if (tm.tm_year >= 1900)
	            tm.tm_year -= 1900;
	
	        break;

            case 3: /* time of day */
	
		if (sscanf (t, "%d:%d:%d", &hour, &min, &sec) == 3)
	            ;
		else if (sscanf (t, "%d:%d", &hour, &min) == 2)
	            sec = 0;
		else {
	            return(-1);
		}

	        tm.tm_hour = hour;
		tm.tm_min = min;
		tm.tm_sec = sec;
		break;

          case 4: /* timezone */
		/* sometimes we see things like (MST) or (-0700) so attempt to
	         * compensate by uncommenting the string if non-RFC822 compliant
	         */
		ptz = utils_uncomment_timezone (tzstr, sizeof (tzstr), t);

		if (*ptz == '+' || *ptz == '-')	{
	  
	    	    if (ptz[1] && ptz[2] && ptz[3] && ptz[4]
		    && isdigit ((unsigned char) ptz[1]) && isdigit ((unsigned char) ptz[2])
	    	    && isdigit ((unsigned char) ptz[3]) && isdigit ((unsigned char) ptz[4])) {
	    
			zhours = (ptz[1] - '0') * 10 + (ptz[2] - '0');
			zminutes = (ptz[3] - '0') * 10 + (ptz[4] - '0');

			if (ptz[0] == '-')
	            	    zoccident = 1;
	            }
		} else {

	            for (i = 0; TimeZones[i].tzname; i++)
			if (!strcasecmp (TimeZones[i].tzname, ptz)){
	                    zhours = TimeZones[i].zhours;
	                    zminutes = TimeZones[i].zminutes;
	                    zoccident = TimeZones[i].zoccident;
	                    break;
			}

		    /* ad hoc support for the European MET (now officially CET) TZ */
	            if (strcasecmp (t, "MET") == 0) {
			if ((t = strtok (NULL, " \t")) != NULL) {
	                    if (!strcasecmp (t, "DST"))
				zhours++;
			}
	            }
		}

		tz_offset = zhours * 3600 + zminutes * 60;
	
		if (!zoccident)
	            tz_offset = -tz_offset;
		break;
	}
	count++;
	t = 0;
    }

    if (count < 4) /* don't check for missing timezone */  {
	return (-1);
    }

    if (h){
	h->zhours = zhours;
	h->zminutes = zminutes;
	h->zoccident = zoccident;
    }

    return (utils_mktime (&tm, 0) + tz_offset);
}

/* extract the first substring that looks like a message-id */
static char *message_extract_message_id (const char *s){

    const char *p;
    char *r;
    size_t l;

    if ((s = strchr (s, '<')) == NULL || (p = strchr (s, '>')) == NULL)
	return (NULL);

    l = (size_t)(p - s) + 1;
    r = g_malloc (l + 1);
    memcpy (r, s, l);
    r[l] = 0;

    return (r);
}

static GList *message_parse_references (char *s){

  GList *list = NULL;

    while ((s = strtok (s, " \t")) != NULL) {
    /*
     * some mail clients add other garbage besides message-ids, so do a quick
     * check to make sure this looks like a valid message-id
     */
	if (*s == '<') {
	    list = g_list_append(list, g_strdup(s));
	}

       s = NULL;
  }

  return list;
}

static void message_parse_content_disposition (char *s, Body *content){

    GList *params;

    if (!strncasecmp ("inline", s, 6))
	content->disposition = DISPINLINE;
    else if (!strncasecmp ("form-data", s, 9))
        content->disposition = DISPFORMDATA;
    else
	content->disposition = DISPATTACH;

    /* Check to see if a default filename was given */
    if ((s = strchr (s, ';')) != NULL) {
	s++;
	SKIPWS (s);

	params = parameters_parse(s);

	if (params){
	
	    if ((s = parameter_get ("filename", params)) != NULL) {
	        G_FREE(content->filename);
		content->filename = g_strdup(s); 
    	    }

	    if ((s = parameter_get ("name", params)) != NULL)
        	content->form_name = g_strdup (s);
	
	    parameters_free(params);
	}	    
    }
}


static void message_parse_content_type (char *s, Body *content) {

    char *pc;
    char *subtype;

    G_FREE(content->subtype);
    parameters_free(content->parameter);

    /* First extract any existing parameters */
    if ((pc = strchr(s, ';')) != NULL){
	*pc++ = 0;

	while (*pc && ISSPACE (*pc))
            pc++;
    
	content->parameter = parameters_parse(pc);

        /* Some pre-RFC1521 gateways still use the "name=filename" convention */
	if ((pc = parameter_get("name", content->parameter)) != NULL)
            content->filename = g_strdup(pc);
    }
  
    /* Now get the subtype */
    if ((subtype = strchr(s, '/'))) {

	*subtype++ = '\0';
	for(pc = subtype; *pc && !ISSPACE(*pc) && *pc != ';'; pc++)
            ;

	*pc = '\0';
	content->subtype = g_strdup(subtype);
    }

    /* Finally, get the major type */
    content->type = message_check_mime_type(s);

    if (content->type == TYPEOTHER){
       content->xtype = g_strdup(s);
    }
  
    if (content->subtype == NULL) {
	/* Some older non-MIME mailers (i.e., mailtool, elm) have a content-type
         * field, so we can attempt to convert the type to BODY here.
         */
	if (content->type == TYPETEXT)
            content->subtype = g_strdup("plain");
	else if (content->type == TYPEAUDIO)
            content->subtype = g_strdup("basic");
	else if (content->type == TYPEMESSAGE)
            content->subtype = g_strdup("rfc822");
	else if (content->type == TYPEOTHER) {
          
	    char buffer[SHORT_STRING];

            content->type = TYPEAPPLICATION;
            snprintf (buffer, sizeof(buffer), "x-%s", s);
            content->subtype = g_strdup(buffer);
	} else
            content->subtype = g_strdup("x-unknown");
    }

}


int message_check_encoding (const char *c){

    if (strncasecmp ("7bit", c, sizeof ("7bit")-1) == 0)
	return (ENC7BIT);
    else if (strncasecmp ("8bit", c, sizeof ("8bit")-1) == 0)
	return (ENC8BIT);
    else if (strncasecmp ("binary", c, sizeof ("binary")-1) == 0)
	return (ENCBINARY);
    else if (strncasecmp ("quoted-printable", c, sizeof ("quoted-printable")-1) == 0)
	return (ENCQUOTEDPRINTABLE);
    else if (strncasecmp ("base64", c, sizeof("base64")-1) == 0)
	return (ENCBASE64);
    else
        return (ENCOTHER);
}

int message_check_mime_type (const char *s){

    if (strcasecmp ("text", s) == 0)
	return TYPETEXT;
    else if (strcasecmp ("multipart", s) == 0)
	return TYPEMULTIPART;
    else if (strcasecmp ("application", s) == 0)
	return TYPEAPPLICATION;
    else if (strcasecmp ("message", s) == 0)
	return TYPEMESSAGE;
    else if (strcasecmp ("image", s) == 0)
	return TYPEIMAGE;
    else if (strcasecmp ("audio", s) == 0)
	return TYPEAUDIO;
    else if (strcasecmp ("video", s) == 0)
	return TYPEVIDEO;
    else
	return TYPEOTHER;
}


void message_parse_mime_message (Message *msg){

    Mailbox *mbox;

    if (msg->content->type != TYPEMESSAGE && msg->content->type != TYPEMULTIPART){
	return; /* nothing to do */
    }

    mbox = msg->mailbox;
    
    if (mbox){
	fseek (mbox->fp, msg->content->offset, 0);
    
	if (msg->content->type == TYPEMULTIPART) {
            if (!msg->content->parts)
		msg->content->parts = message_parse_multipart( mbox->fp, parameter_get("boundary", msg->content->parameter), 
							    msg->content->offset + msg->content->length, 
							    strcasecmp ("digest", msg->content->subtype) == 0);
	} else {
	    if (!msg->content->parts)
		msg->content->parts = message_parse_message(mbox->fp, msg->content);
	}

	/* try to recover from parsing error */
	if (!msg->content->parts){
	    msg->content->type = TYPETEXT;
    	    g_free(msg->content->subtype);
	    msg->content->subtype = g_strdup("plain");
	}
    }
}

GList *message_parse_multipart(FILE *fp, const char *boundary, long end_off, int digest){

    int 	blen, len, crlf = 0;
    char 	buffer[LONG_STRING];
    Body 	*body, *last = NULL;
    GList 	*l, *list = NULL;
    int 	i;
    int 	final = 0; /* did we see the ending boundary? */

    if (!boundary) {
	/* FIXME: display error to user */
	dprint(1, (stderr,"message_parse_multipart(): multipart message has no boundary specified\n")); 
	return (NULL);
    }

    blen = strlen (boundary);

    while (ftell (fp) < end_off && fgets (buffer, LONG_STRING, fp) != NULL) {

	len = strlen (buffer);

        /* take note of the line ending.  I'm assuming that either all endings
         * will use <CR><LF> or none will.
         */

	if (len > 1 && buffer[len - 2] == '\r')
            crlf = 1;

        if (buffer[0] == '-' && buffer[1] == '-' && strncmp (buffer + 2, boundary, blen) == 0) {

            if (last) {

		last->length = ftell (fp) - last->offset - len - 1 - crlf;
		
		if (last->parts){
		    Body *tmp = (Body *)(last->parts->data);
		    if (tmp->length == 0)
	        	tmp->length = ftell (fp) - tmp->offset - len - 1 - crlf;
		}

		/* if the body is empty, we can end up with a -1 length */
		if (last->length < 0)
	            last->length = 0;
            }

            /* Remove any trailing whitespace, up to the length of the boundary */
            for (i = len - 1; ISSPACE (buffer[i]) && i >= blen + 2; i--)
    		buffer[i] = 0;

            /* Check for the end boundary */
            if (strcmp (buffer + blen + 2, "--") == 0){
		final = 1;
		break; /* done parsing */

            } else if (buffer[2 + blen] == 0) {
		body = message_read_mime_header (fp, digest);

		list = g_list_append(list, body);
		last = body;

            }
	}
    }



    /* in case of missing end boundary, set the length to something reasonable */
    if (last && last->length == 0 && !final)
	last->length = end_off - last->offset;

    /* parse recursive MIME parts */
    l = g_list_first(list);

//    dprint(99, (stderr, "parse1\n"));
    while (l){

	body = (Body *)(l->data);
	
	switch (body->type){

            case TYPEMULTIPART:
		fseek (fp, body->offset, 0);
		body->parts = message_parse_multipart( fp, parameter_get("boundary", body->parameter), 
						    body->offset + body->length, strcasecmp ("digest", body->subtype) == 0);
		break;

            case TYPEMESSAGE:
		if (body->subtype && (strcasecmp (body->subtype, "rfc822") == 0 || strcasecmp (body->subtype, "news") == 0)){
	    	    fseek (fp, last->offset, 0);
	            body->parts = message_parse_message(fp, body);
		}
		break;

	}

	l = g_list_next(l);
    }
//    dprint(99, (stderr, "parseX\n"));
    
    return (g_list_first(list));
}

Body *message_read_mime_header (FILE *fp, int digest){

    Body *p = body_new();
    char *c;
    char *line = g_malloc(LONG_STRING);
    size_t linelen = LONG_STRING;
  
    p->hdr_offset = ftell(fp);

    p->encoding = ENC7BIT; /* default from RFC1521 */
    p->type = digest ? TYPEMESSAGE : TYPETEXT;

    while (*(line = message_read_line (fp, line, &linelen)) != 0){

        /* Find the value of the current header */
	if ((c = strchr (line, ':'))) {
            *c = 0;
            c++;
            SKIPWS (c);
            if (!*c) {
		dprint (1, (stderr, "message_read_mime_header(): skipping empty header field: %s\n", line));
		continue;
            }
	} else {
    	    dprint (1, (stderr, "message_read_mime_header(): bogus MIME header: %s\n", line));
            break;
	}

	if (!strncasecmp ("content-", line, 8)){
            if (!strcasecmp ("type", line + 8))
		message_parse_content_type (c, p);
            else if (!strcasecmp ("transfer-encoding", line + 8))
		p->encoding = message_check_encoding (c);
            else if (!strcasecmp ("disposition", line + 8))
		message_parse_content_disposition (c, p);
            else if (!strcasecmp ("description", line + 8)){
		g_free(p->description);
		p->description = g_strdup(c);
		rfc2047_decode(p->description, p->description, strlen (p->description) + 1);
            }
	}
    }

    p->offset = ftell(fp); /* Mark the start of the real data */

    if (p->type == TYPETEXT && !p->subtype)
	p->subtype = g_strdup ("plain");
    else if (p->type == TYPEMESSAGE && !p->subtype)
	p->subtype = g_strdup ("rfc822");

    g_free(line);	

    return (p);
}

GList *message_parse_message(FILE *fp, Body *parent){

    /* FIXME */
    return NULL;

}

void message_decode_attachment (Body *b, State *s){

    fseek (s->fpin, b->offset, 0);

    switch (b->encoding){

	case ENCQUOTEDPRINTABLE:
            utils_decode_quoted (s, b->length, message_is_text_type (b->type, b->subtype));
            break;

	case ENCBASE64:
            utils_decode_base64 (s, b->length, message_is_text_type (b->type, b->subtype));
            break;

	default:
            utils_decode_xbit (s, b->length, message_is_text_type (b->type, b->subtype));
            break;
  }
}

static int message_is_text_type (int t, char *s){

    if (t == TYPETEXT)
	return 1;

    if (t == TYPEMESSAGE){
	if (!strcasecmp ("delivery-status", s))
        return 1;
    }

    return 0;
}

int message_check_read_status(FILE *file){

    long 	current_loc;
    char 	*line = g_malloc (LONG_STRING);
    char 	*p;
    int 	result = 0;
    size_t 	linelen = LONG_STRING;

    current_loc = ftell(file);

    while (*(line = message_read_line (file, line, &linelen)) != 0) {

	if ((p = strpbrk (line, ": \t")) == NULL || *p != ':') {

            if (strncmp (">From ", line, 6) == 0) {
		current_loc = ftell(file);
	        continue; // just ignore 
            }

            fseek (file, current_loc, 0);
            break; // end of header
	}

	*p = 0;
	p++;
	SKIPWS (p);

	if (!*p){
	    current_loc = ftell(file);
            continue; // skip empty header fields
	}

	if (!strcasecmp (line, "status")) {

	    while (*p){

		if (*p == 'R'){
		    result = 1;
		    break;
		}		    

                p++;

	    }
  
	}	        

	current_loc = ftell(file);
    }

    g_free(line);
    
    return result;
}