/*
 * Glacier - another Gnome email client.
 * Copyright (C) 1999 Lubomir Gelo
 *
 * Glacier uses code from other GPL-ed projects. See THANKS for 
 * credits and copyright info.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#include "includes.h"

const char *Weekdays[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
const char *WeekdayNames[] = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday"};

const char *Months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec", "ERR" };
const char *MonthNames[] = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December", "ERR" };


static sigset_t Sigset;

const char *utils_get_next_word (const char *s){

    while (*s && !ISSPACE (*s))
	s++;

    SKIPWS (s);

    return s;
}

int utils_check_month (const char *s){

    int i;

    for (i = 0; i < 12; i++)
	if (strncasecmp (s, Months[i], 3) == 0)
            return (i);

    return (-1); /* error */
}

int utils_is_day_name (const char *s){

    int i;

    if (!ISSPACE (*(s+3)))
	return 0;

    for (i=0; i<7; i++)
	if (strncasecmp (s, Weekdays[i], 3) == 0)
    	    return 1;
    return 0;
}

int utils_lock_file (const char *path, int fd, int excl, int dot, int timeout){

#if defined (USE_FCNTL) || defined (USE_FLOCK)
    int count;
    int attempt;
    struct stat prev_sb;
#endif
    int r = 0;

#ifdef USE_FCNTL

    struct flock lck;

    count = 0;
    attempt = 0;
  
    memset (&lck, 0, sizeof (struct flock));
    lck.l_type = excl ? F_WRLCK : F_RDLCK;
    lck.l_whence = SEEK_SET;

    while (fcntl (fd, F_SETLK, &lck) == -1) {
    
	struct stat sb;
	fprintf(stderr, "utils_lock_file(): fcntl errno %d.\n", errno);

	if (errno != EAGAIN && errno != EACCES) {
            return (-1);
	}

	if (fstat (fd, &sb) != 0)
	    sb.st_size = 0;
    
	if (count == 0)
            prev_sb = sb;

        /* only unlock file if it is unchanged */
        if (prev_sb.st_size == sb.st_size && ++count >= (timeout ? MAXLOCKATTEMPT : 0)) {
      
    	    if (timeout)
		fprintf (stderr, "Timeeout exceeded while attempting fcntl lock!");
    
            return (-1);
	}

	prev_sb = sb;

	sleep (1);
    }
#endif /* USE_FCNTL */

#ifdef USE_FLOCK
    count = 0;
    attempt = 0;

    while (flock (fd, (excl ? LOCK_EX : LOCK_SH) | LOCK_NB) == -1) {

	struct stat sb;

        if (errno != EWOULDBLOCK) {
	    r = -1;
    	    break;
	}

	if (fstat(fd,&sb) != 0 )
    	    sb.st_size=0;
    
	if (count == 0)
    	    prev_sb=sb;

	/* only unlock file if it is unchanged */
	if (prev_sb.st_size == sb.st_size && ++count >= (timeout?MAXLOCKATTEMPT:0)) {

    	    if (timeout)
		fprintf(stderr, "Timeout exceeded while attempting flock lock!");

	    r = -1;
            break;
	}

	prev_sb = sb;

	sleep (1);
    }
#endif /* USE_FLOCK */

    if (r == -1){
	/* release any other locks obtained in this routine */

#ifdef USE_FCNTL
	lck.l_type = F_UNLCK;
	fcntl (fd, F_SETLK, &lck);
#endif /* USE_FCNTL */

#ifdef USE_FLOCK
	flock (fd, LOCK_UN);
#endif /* USE_FLOCK */

	return (-1);
    }

    return 0;
}

int utils_unlock_file (const char *path, int fd){

#ifdef USE_FCNTL
    struct flock unlockit = { F_UNLCK, 0, 0, 0 };

    memset (&unlockit, 0, sizeof (struct flock));
    unlockit.l_type = F_UNLCK;
    unlockit.l_whence = SEEK_SET;
    fcntl (fd, F_SETLK, &unlockit);
#endif

#ifdef USE_FLOCK
    flock (fd, LOCK_UN);
#endif

    return 0;
}

/* signals which are important to block while doing critical ops */
void utils_block_signals (void){

    if (!glacier_app.signals_blocked){
	sigemptyset (&Sigset);
	sigaddset (&Sigset, SIGINT);
	sigaddset (&Sigset, SIGWINCH);
	sigaddset (&Sigset, SIGHUP);
	sigaddset (&Sigset, SIGTERM);
	sigaddset (&Sigset, SIGTSTP);
	sigprocmask (SIG_BLOCK, &Sigset, 0);
	glacier_app.signals_blocked = TRUE;
    }
}

/* restore the previous signal mask */
void utils_unblock_signals (void){

    if (glacier_app.signals_blocked){
	sigprocmask (SIG_UNBLOCK, &Sigset, 0);
	glacier_app.signals_blocked = FALSE;
    }
}



/* returns the seconds west of UTC given `g' and its corresponding gmtime() representation */
time_t utils_compute_tz (time_t g, struct tm *utc){

    struct tm *lt = localtime (&g);
    time_t t;

    t = (((utc->tm_hour - lt->tm_hour) * 60) + (utc->tm_min - lt->tm_min)) * 60;

    switch (utc->tm_yday - lt->tm_yday){

	case 0:
    	    break;

	case 1:
	case -364:
	case -365:
            t += 24 * 60 * 60;
            break;

	case -1:
	case 364:
	case 365:
            t -= 24 * 60 * 60;
            break;

	default:
            fprintf(stderr,"Error in utils_compute_tz\n");
    }

    return t;
}

const char *utils_uncomment_timezone (char *buf, size_t buflen, const char *tz){

    char *p;
    size_t len;

    if (*tz != '(')
	return tz; /* no need to do anything */

    tz++;
    SKIPWS (tz);

    if ((p = strpbrk (tz, " )")) == NULL)
	return tz;

    len = p - tz;
    if (len > buflen - 1)
	len = buflen - 1;

    memcpy (buf, tz, len);
    buf[len] = 0;

    return buf;
}


/* 
    Today  	13:10
    Friday	19:20
    10 Jul	18:20
    Wednsday
    10 Jul 99	13:30
*/

void utils_print_friendly_time(time_t ptime, gchar **string_date, gchar **string_time){

    gchar	string[STRING];
    time_t 	now;
    struct tm 	*tm;
    
    now = time(NULL);

    tm = gmtime(&ptime);
    sprintf(string, "%02d:%02d", tm->tm_hour, tm->tm_min);
    *string_time = g_strdup(string);
    
    if (ptime + SECONDS_PER_DAY > now){
	*string_date = g_strdup( _("Today"));
    } else if (ptime + SECONDS_PER_WEEK > now){
	*string_date = g_strdup(WeekdayNames[tm->tm_wday]); 
	/* day of the week */
    } else {
	sprintf(string, "%d %s %d", tm->tm_mday, Months[tm->tm_mon], tm->tm_year);
	*string_date = g_strdup(string);
    }

}

gchar *utils_print_time(time_t ptime){

    gchar	string[STRING];
    struct tm 	*tm;

    tm = gmtime(&ptime);
    sprintf(string, _("%s, %d %s %d %02d:%02d:%02d"), WeekdayNames[tm->tm_wday], tm->tm_mday, MonthNames[tm->tm_mon], tm->tm_year,
		    tm->tm_hour, tm->tm_min, tm->tm_sec);

    return(g_strdup(string));
}


time_t utils_local_tz (void){

    struct tm *ptm;
    struct tm utc;
    time_t now;

    now = time (NULL);
    ptm = gmtime (&now);

    /* need to make a copy because gmtime/localtime return a pointer to static memory (grr!) */
    memcpy (&utc, ptm, sizeof (utc));

    return (utils_compute_tz (now, &utc));
}

/* converts struct tm to time_t, but does not take the local timezone into account unless ``local'' is nonzero */
time_t utils_mktime (struct tm *t, int local){

    time_t g;

    static int AccumDaysPerMonth[12] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334 };

    /* Compute the number of days since January 1 in the same year */
    g = AccumDaysPerMonth [t->tm_mon % 12];

    /* The leap years are 1972 and every 4. year until 2096,* but this algoritm will fail after year 2099 */
    g += t->tm_mday;

    if ((t->tm_year % 4) || t->tm_mon < 2)
	g--;

    t->tm_yday = g;

    /* Compute the number of days since January 1, 1970 */
    g += (t->tm_year - 70) * 365;
    g += (t->tm_year - 69) / 4;

    /* Compute the number of hours */
    g *= 24;
    g += t->tm_hour;

    /* Compute the number of minutes */
    g *= 60;
    g += t->tm_min;

    /* Compute the number of seconds */
    g *= 60;
    g += t->tm_sec;

    if (local)
	g += utils_compute_tz (g, t);

    return (g);
}

void utils_normalize_time (struct tm *tm){

    static char DaysPerMonth[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

    while (tm->tm_sec < 0) {
	tm->tm_sec += 60;
	tm->tm_min--;
    }

    while (tm->tm_min < 0) {
	tm->tm_min += 60;
	tm->tm_hour--;
    }

    while (tm->tm_hour < 0) {
	tm->tm_hour += 24;
	tm->tm_mday--;
    }

    while (tm->tm_mon < 0) {
	tm->tm_mon += 12;
	tm->tm_year--;
    }

    while (tm->tm_mday < 0) {
	if (tm->tm_mon)
            tm->tm_mon--;
	else  {
            tm->tm_mon = 11;
            tm->tm_year--;
	}
	tm->tm_mday += DaysPerMonth[tm->tm_mon];
    }
}

void utils_dequote_comment (char *s){

    char *w = s;

    for (; *s; s++){
	if (*s == '\\'){
    	    if (!*++s)
		break; /* error? */
      
	    *w++ = *s;
	} else if (*s != '\"') {
            if (w != s)
		*w = *s;

            w++;
	}
    }

    *w = 0;
}

char *utils_substrdup (const char *begin, const char *end){

    size_t len;
    char *p;

    len = end - begin;
    p = g_malloc (len + 1);
    memcpy (p, begin, len);
    p[len] = 0;

    return p;
}

void utils_gui_update_hook (int count, int total){

    while (gtk_events_pending ())
        gtk_main_iteration ();
}

size_t utils_readfile (FILE * fp, char **buf) {

    size_t size;
    off_t offset;
    int r;
    int fd = fileno (fp);
    struct stat statbuf;

    if (fstat (fd, &statbuf) == -1)
	return -1;

    size = statbuf.st_size;

    if (!size) {
        *buf = NULL;
        return size;
    }

    lseek (fd, 0, SEEK_SET);

    *buf = (char *) g_malloc (size + 1);

    if (*buf == NULL) {
	return -1;
    }

    offset = 0;

    while (offset < size) {
        r = read (fd, *buf + offset, size - offset);
	if (r == 0)
	    return offset;

        if (r > 0){
	    offset += r;
	} else if ((errno != EAGAIN) && (errno != EINTR)){
	    perror ("Error reading file:");
	    return -1;
	}
    }

    return(size + 1);
}

FILE *utils_mktemp (char **filename){

    int 	fd;

    do {
	*filename = tempnam (NULL, "g");
	fd = open (*filename, O_CREAT | O_EXCL | O_TRUNC | O_RDWR, 0600);
	if (fd == -1)
	    free(*filename);
    } while (fd == -1);

    return(fdopen(fd, "w+"));
}



void utils_decode_xbit (State *s, long len, int istext){

    int c;
    int lbreak = 1;
  
    if (istext){

	while ((c = fgetc(s->fpin)) != EOF && len--){

    	    if (lbreak && s->prefix) {
		fputs(s->prefix, s->fpout);
		lbreak = 0;
            }
	  
            if (c == '\r' && len) {

		int ch;
	
		if((ch = fgetc(s->fpin)) != '\n')
	            ungetc(ch, s->fpin);
		else {
	            c = ch;
	            len--;
		}
            }
      
            fputc(c, s->fpout);
      
            if(c == '\n')
		lbreak = 1;
	}
    } else
	utils_copy_bytes (s->fpin, s->fpout, len);
}

void utils_decode_quoted (State *s, long len, int istext){

    int ch, lbreak = 1;

    while (len > 0){

	if ((ch = fgetc(s->fpin)) == EOF)
            break;

	len--;
    
	if (s->prefix && lbreak)
            fputs(s->prefix, s->fpout);
    
	lbreak = 0;

	if (ch == '='){
            int ch1, ch2;
      
	    if(!len || (ch1 = fgetc(s->fpin)) == EOF)
		break;

            len--;
      
            if (ch1 == '\n' || ch1 == '\r' || ch1 == ' ' || ch1 == '\t') {

		/* Skip whitespace at the end of the line since MIME does not allow for it */
		if(ch1 != '\n'){

	            while(len && (ch1 = fgetc(s->fpin)) != EOF){
			len--;
			if(ch1 == '\n')
	                    break;
	            }
		}

		if(ch1 == EOF)
	            break;

		ch = EOF;

	    } else {

	        if(!len || (ch2 = fgetc(s->fpin)) == EOF)
	            break;
		
	        len--;
	
    		ch = hexval (ch1) << 4;
    		ch |= hexval (ch2);
            }

	} else if (istext && ch == '\r') {

            int ch1;

            if((ch1 =fgetc(s->fpin)) == '\n'){
		ch = ch1;
		len--;
            } else
		ungetc(ch1, s->fpin);
	}

	if(ch != EOF)
              fputc(ch, s->fpout);

	if(ch == '\n')
              lbreak = 1;
    }
}

void utils_decode_base64 (State *s, long len, int istext){

    char buf[5];
    int c1, c2, c3, c4, ch, cr = 0, i;

    buf[4] = 0;

    if (s->prefix) 
	fputs(s->prefix, s->fpout);

    while (len > 0){
	
	for (i = 0 ; i < 4 && len > 0 ; len--){

            if ((ch = fgetc (s->fpin)) == EOF)
		return;

    	    if (!ISSPACE (ch))
		buf[i++] = ch;
	}

	if (i != 4)
            return; /* didn't get a multiple of four chars! */

        c1 = base64val (buf[0]);
        c2 = base64val (buf[1]);
        ch = (c1 << 2) | (c2 >> 4);
	
        if (cr && ch != '\n') 
          fputc('\r', s->fpout);

	cr = 0;
      
        if (istext && ch == '\r')
            cr = 1;
	else {
            fputc(ch, s->fpout);
            if (ch == '\n' && s->prefix) 
		fputs(s->prefix, s->fpout);
	}

	if (buf[2] == '=')
            break;

	c3 = base64val (buf[2]);
	ch = ((c2 & 0xf) << 4) | (c3 >> 2);

	if (cr && ch != '\n')
            fputc('\r', s->fpout);

	cr = 0;

	if (istext && ch == '\r')
            cr = 1;
	else {
            fputc(ch, s->fpout);

            if (ch == '\n' && s->prefix)
		fputs(s->prefix, s->fpout);
	}

	if (buf[3] == '=') break;
	c4 = base64val (buf[3]);
	ch = ((c3 & 0x3) << 6) | c4;

	if (cr && ch != '\n')
            fputc('\r', s->fpout);

	cr = 0;

	if (istext && ch == '\r')
            cr = 1;
	else {
            fputc(ch, s->fpout);

            if (ch == '\n' && s->prefix)
		fputs(s->prefix, s->fpout);
	}
    }
}    

int utils_copy_bytes (FILE *in, FILE *out, size_t size){

    char buf[2048];
    size_t chunk;

    while (size > 0){

	chunk = (size > sizeof (buf)) ? sizeof (buf) : size;

	if ((chunk = fread (buf, 1, chunk, in)) < 1)
            break;

	if (fwrite (buf, 1, chunk, out) != chunk) {
            return (-1);
	}

	size -= chunk;
    }

    return 0;
}
